from autofit.tools import path_util
from autofit.optimize import non_linear as nl
from autofit.mapper import prior
from autofit.mapper import prior_model
from autocti.data import mask as msk
from autocti.pipeline import pipeline as pl
from autocti.pipeline import phase as ph
from autocti.model import arctic_params

# In this pipeline, we'll perform an analysis which fits two parallel trap species to a set of charge
# injection imaging data. This will include a hyper-phase which scales the noise in the analysis, to prevent
# over-fitting the highest S/N charge injection images. The pipeline uses four phases:

# Phase 1) Fit a small section (60 columns of every charge injection) using a parallel CTI model
#          with 1 trap species and a model for the parallel CCD volume filling parameters.

# Phase 2) Fit a small section (again, 60 columns) using a parallel CTI model with 2 trap species and a model for the
#          parallel CCD volume filling parameters. The priors on trap densities and volume filling parameters are
#          initialized from the results of phase 1.

# Phase 3) Use the best-fit model from phase 2 to scale the noise of each image, to ensure that the higher and
#          lower S/N images are weighted more equally in their contribution to the likelihood.

# Phase 4) Refit the phase 2 model, using priors initialized from the results of phase 2 and the scaled noise-map
#          computed in phase 3.

def make_pipeline(phase_folders=None, tag_phases=True, mask_function=msk.Mask.empty_for_shape,
                  parallel_front_edge_mask_rows=None, parallel_trails_mask_rows=None,
                  serial_front_edge_mask_columns=None, serial_trails_mask_columns=None,
                  parallel_total_density_range=None, serial_total_density_range=None,
                  cosmic_ray_parallel_buffer=None, cosmic_ray_serial_buffer=None, cosmic_ray_diagonal_buffer=None):

    pipeline_name = 'pipeline__hyper_parallel_x2_serial'

    # This function uses the phase folders and pipeline name to set up the output directory structure,
    # e.g. 'autolens_workspace/output/phase_folder_1/phase_folder_2/pipeline_name/phase_name/'
    phase_folders = path_util.phase_folders_from_phase_folders_and_pipeline_name(phase_folders=phase_folders,
                                                                                pipeline_name=pipeline_name)

    ### PHASE 1 ###

    # In phase 1, we will fit the data with a two species parallel and three species serial CTI model. In this
    # phase we will:

    # 1) Initialize the priors on this model from phases 2 and 4 of the pipeline.

    class ParallelSerialPhase(ph.ParallelSerialPhase):

        def pass_priors(self, results):
            
            self.parallel_ccd = results.from_phase('phase_2_parallel_x2').variable.parallel_ccd
            self.parallel_species = results.from_phase('phase_2_parallel_x2').variable.parallel_species
            self.parallel_ccd.well_fill_alpha = 1.0
            self.parallel_ccd.well_fill_gamma = 0.0
            self.serial_ccd = results.from_phase('phase_4_serial_x3').variable.serial_ccd
            self.serial_species = results.from_phase('phase_4_serial_x3').variable.serial_species
            self.serial_ccd.well_fill_alpha = 1.0
            self.serial_ccd.well_fill_gamma = 0.0

    phase1 = ParallelSerialPhase(phase_name='phase_1_parallel_x2_serial_x3', 
                                 phase_folders=phase_folders,
                                 tag_phases=tag_phases,
                                 parallel_species=[prior_model.PriorModel(arctic_params.Species),
                                                   prior_model.PriorModel(arctic_params.Species)],
                                 parallel_ccd=arctic_params.CCD,
                                 serial_species=[prior_model.PriorModel(arctic_params.Species),
                                                 prior_model.PriorModel(arctic_params.Species),
                                                 prior_model.PriorModel(arctic_params.Species)],
                                 serial_ccd=arctic_params.CCD,
                                 optimizer_class=nl.MultiNest, mask_function=mask_function,
                                 parallel_front_edge_mask_rows=parallel_front_edge_mask_rows,
                                 parallel_trails_mask_rows=parallel_trails_mask_rows,
                                 serial_front_edge_mask_columns=serial_front_edge_mask_columns,
                                 serial_trails_mask_columns=serial_trails_mask_columns,
                                 parallel_total_density_range=parallel_total_density_range,
                                 serial_total_density_range=serial_total_density_range,
                                 cosmic_ray_parallel_buffer=cosmic_ray_parallel_buffer,
                                 cosmic_ray_serial_buffer=cosmic_ray_serial_buffer,
                                 cosmic_ray_diagonal_buffer=cosmic_ray_diagonal_buffer)

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 80
    phase1.optimizer.sampling_efficiency = 0.2

    ### PHASE 2 ###

    # The best fit model of phase 2 is used to create a 'noise-scaling' map for every charge injection image. These
    # noise-scaling maps are used in conjunction with 'hyper-noise' models to scale the noise-maps in a way that
    # equally weights the fit across all charge injection images.

    class ParallelSerialHyperModelFixedPhase(ph.ParallelSerialHyperPhase):

        def pass_priors(self, results):

            self.parallel_species = \
                results.from_phase('phase_1_parallel_x2_serial_x3').constant.parallel_species
            self.parallel_ccd = \
                results.from_phase('phase_1_parallel_x2_serial_x3').constant.parallel_ccd
            self.serial_species = \
                results.from_phase('phase_1_parallel_x2_serial_x3').constant.serial_species
            self.serial_ccd = \
                results.from_phase('phase_1_parallel_x2_serial_x3').constant.serial_ccd

    phase2 = ParallelSerialHyperModelFixedPhase(phase_name='phase_2_noise_scaling', phase_folders=phase_folders,
                                                tag_phases=tag_phases,
                                                parallel_species=[prior_model.PriorModel(arctic_params.Species),
                                                                  prior_model.PriorModel(arctic_params.Species)],
                                                parallel_ccd=arctic_params.CCD,
                                                serial_species=[prior_model.PriorModel(arctic_params.Species),
                                                                prior_model.PriorModel(arctic_params.Species),
                                                                prior_model.PriorModel(arctic_params.Species)],
                                                serial_ccd=arctic_params.CCD,
                                                optimizer_class=nl.MultiNest, mask_function=mask_function,
                                                parallel_front_edge_mask_rows=parallel_front_edge_mask_rows,
                                                parallel_trails_mask_rows=parallel_trails_mask_rows,
                                                serial_front_edge_mask_columns=serial_front_edge_mask_columns,
                                                serial_trails_mask_columns=serial_trails_mask_columns,
                                                parallel_total_density_range=parallel_total_density_range,
                                                serial_total_density_range=serial_total_density_range,
                                                cosmic_ray_parallel_buffer=cosmic_ray_parallel_buffer,
                                                cosmic_ray_serial_buffer=cosmic_ray_serial_buffer,
                                                cosmic_ray_diagonal_buffer=cosmic_ray_diagonal_buffer)

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.nr_live_points = 30
    phase2.optimizer.sampling_efficiency = 0.2

    ### PHASE 3 ###

    # In phase 3, we will fit the data with a 2 species parallel and 3 species serial CTI model. In this
    # phase we will:

    # 1) Use the scaled noise-map computed in phase 2.
    # 2) Initialize the priors on the parallel and serial CTI models from the results of phase 1.

    class ParallelSerialHyperFixedPhase(ph.ParallelSerialHyperPhase):

        def pass_priors(self, results):

            self.hyper_noise_scalars = results.from_phase('phase_2_noise_scaling').constant.hyper_noise_scalars

            self.parallel_species = \
                results.from_phase('phase_1_parallel_x2_serial_x3').variable.parallel_species
            self.parallel_ccd = \
                results.from_phase('phase_1_parallel_x2_serial_x3').variable.parallel_ccd
            self.parallel_ccd.well_fill_alpha = 1.0
            self.parallel_ccd.well_fill_gamma = 0.0

            self.serial_species = \
                results.from_phase('phase_1_parallel_x2_serial_x3').variable.serial_species
            self.serial_ccd = \
                results.from_phase('phase_1_parallel_x2_serial_x3').variable.serial_ccd
            self.serial_ccd.well_fill_alpha = 1.0
            self.serial_ccd.well_fill_gamma = 0.0

    phase3 = ParallelSerialHyperFixedPhase(phase_name='phase_3_parallel_x2_serial_x3_species', phase_folders=phase_folders,
                                           tag_phases=tag_phases,
                                           parallel_species=[prior_model.PriorModel(arctic_params.Species),
                                                             prior_model.PriorModel(arctic_params.Species)],
                                           parallel_ccd=arctic_params.CCD,
                                           serial_species=[prior_model.PriorModel(arctic_params.Species),
                                                           prior_model.PriorModel(arctic_params.Species),
                                                           prior_model.PriorModel(arctic_params.Species)],
                                           serial_ccd=arctic_params.CCD,
                                           optimizer_class=nl.MultiNest, mask_function=mask_function,
                                           parallel_front_edge_mask_rows=parallel_front_edge_mask_rows,
                                           parallel_trails_mask_rows=parallel_trails_mask_rows,
                                           serial_front_edge_mask_columns=serial_front_edge_mask_columns,
                                           serial_trails_mask_columns=serial_trails_mask_columns,
                                           parallel_total_density_range=parallel_total_density_range,
                                           serial_total_density_range=serial_total_density_range,
                                           cosmic_ray_parallel_buffer=cosmic_ray_parallel_buffer,
                                           cosmic_ray_serial_buffer=cosmic_ray_serial_buffer,
                                           cosmic_ray_diagonal_buffer=cosmic_ray_diagonal_buffer)

    # For the final CTI model, constant efficiency mode has a tendency to sample parameter space too fast and infer an
    # inaccurate model. Thus, we turn it off for phase 2.

    phase3.optimizer.const_efficiency_mode = False
    phase3.optimizer.n_live_points = 50
    phase3.optimizer.sampling_efficiency = 0.3

    return pl.Pipeline(pipeline_name, phase1, phase2, phase3)


#
# Copyright (C) 2012-2020 Euclid Science Ground Segment
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3.0 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
File: tests/python/PyArCTIC_test.py

Created on: 02/13/18
Author: James Nightingale
"""

import numpy as np
import pytest

from autocti.model import arctic_params
from autocti.model import arctic_settings
from autocti.model import pyarctic


class TestArcticAddCTI:

    class TestSquareDimensionsCTI:

        def test__horizontal_line__line_loses_charge_trails_appear(self):

            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[2, :] += 100

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0,
                                              well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=True, readout_offset=0)

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[0:2, :] == 0.0).all()  # First four rows should all remain zero
            assert (image_difference[2,
                    :] < 0.0).all()  # All pixels in the charge line should lose charge due to capture
            assert (image_difference[3:-1, :] > 0.0).all()  # All other pixels should have charge trailed into them

        def test__vertical_line__no_trails(self):
            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[:, 2] += 100

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0,
                                              well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=True, readout_offset=0)

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[:, 0:2] == 0.0).all()  # Most pixels unchanged
            assert (image_difference[:, 3:-1] == 0.0).all()
            assert (image_difference[:, 2] < 0.0).all()  # charge line still loses charge

        def test__double_trap_density__more_captures_so_brighter_trails(self):

            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[2, :] += 100

            # SETUP TWO SETS OF PARAMETERS WITH ONE PARAMETER DOUBLED #

            species_0 = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            species_1 = arctic_params.Species(trap_density=0.2, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0,
                                              well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=True, readout_offset=0)

            # NOW GENERATE THE IMAGE POST CTI OF EACH SET

            image_post_cti_0 = pyarctic.call_arctic(image=image_pre_cti, species=[species_0], ccd=ccd,
                                                  settings=settings, correct_cti=False)
            image_post_cti_1 = pyarctic.call_arctic(image=image_pre_cti, species=[species_1], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            assert (image_post_cti_0[0:2, :] == 0.0).all()  # First four rows should all remain zero
            assert (image_post_cti_1[0:2, :] == 0.0).all()  # First four rows should all remain zero
            # noinspection PyUnresolvedReferences
            assert (image_post_cti_0[2, :] > image_post_cti_1[4, :]).all()  # charge line loses less charge in image 1
            # noinspection PyUnresolvedReferences
            assert (image_post_cti_0[3:-1, :] < image_post_cti_1[5:-1, :]).all()
            # fewer pixels trailed behind in image 2

        def test__double_trap_lifetime__longer_release_so_fainter_trails(self):

            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[2, :] += 100

            # SETUP TWO SETS OF PARAMETERS WITH ONE PARAMETER DOUBLED #

            species_0 = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            species_1 = arctic_params.Species(trap_density=0.1, trap_lifetime=2.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0,
                                              well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=True, readout_offset=0)

            # NOW GENERATE THE IMAGE POST CTI OF EACH SET

            image_post_cti_0 = pyarctic.call_arctic(image=image_pre_cti, species=[species_0], ccd=ccd,
                                                  settings=settings, correct_cti=False)
            image_post_cti_1 = pyarctic.call_arctic(image=image_pre_cti, species=[species_1], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            assert (image_post_cti_0[0:2, :] == 0.0).all()  # First four rows should all remain zero
            assert (image_post_cti_1[0:2, :] == 0.0).all()  # First four rows should all remain zero
            # noinspection PyUnresolvedReferences
            assert (image_post_cti_0[2, :] == image_post_cti_1[2, :]).all()  # charge line loses equal amount of charge
            # noinspection PyUnresolvedReferences
            assert (image_post_cti_0[3:-1, :] > image_post_cti_1[3:-1, :]).all()
            # each trail in pixel 2 is 'longer' so fainter

        def test__increase_beta__fewer_captures_fainter_trail(self):

            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[2, :] += 100

            # SETUP TWO SETS OF PARAMETERS WITH ONE PARAMETER DOUBLED #

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd_0 = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0,
                                               well_fill_beta=0.8, well_fill_gamma=0.0)
            ccd_1 = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0,
                                               well_fill_beta=0.9, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=True, readout_offset=0)

            # NOW GENERATE THE IMAGE POST CTI OF EACH SET

            image_post_cti_0 = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd_0,
                                                  settings=settings, correct_cti=False)
            image_post_cti_1 = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd_1,
                                                  settings=settings, correct_cti=False)

            assert (image_post_cti_0[0:2, :] == 0.0).all()  # First four rows should all remain zero
            assert (image_post_cti_1[0:2, :] == 0.0).all()  # First four rows should all remain zero
            # noinspection PyUnresolvedReferences
            assert (image_post_cti_0[2, :] < image_post_cti_1[2, :]).all()
            # charge line loses less charge with higher beta
            # noinspection PyUnresolvedReferences
            assert (image_post_cti_0[3:-1, :] > image_post_cti_1[3:-1, :]).all()  # so less electrons trailed into image

        def test__two_traps_half_density_of_one__same_trails(self):

            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[2, :] += 100

            # SETUP TWO SETS OF PARAMETERS WITH ONE PARAMETER DOUBLED #

            species_0 = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            species_1 = arctic_params.Species(trap_density=0.05, trap_lifetime=1.0)

            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0,
                                              well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=True, readout_offset=0)

            # NOW GENERATE THE IMAGE POST CTI OF EACH SET

            image_post_cti_0 = pyarctic.call_arctic(image=image_pre_cti, species=[species_0], ccd=ccd,
                                                  settings=settings, correct_cti=False)
            image_post_cti_1 = pyarctic.call_arctic(image=image_pre_cti,
                                                    species=[species_1, species_1], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            # noinspection PyUnresolvedReferences
            assert (image_post_cti_0 == image_post_cti_1).all()

        def test__delta_functions__add_cti_only_behind_them(self):

            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[1, 1] += 100  # Delta 1
            image_pre_cti[3, 3] += 100  # Delta 2
            image_pre_cti[2, 4] += 100  # Delta 3

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0,
                                              well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=True, readout_offset=0)

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[:, 0] == 0.0).all()  # No Delta, no charge

            assert (image_difference[0, 1] == 0.0)  # No charge in front of Delta 1
            assert (image_difference[1, 1] < 0.0)  # Delta 1 loses charge
            assert (image_difference[2:5, 1] > 0.0).all()  # Delta 1 trails

            assert (image_difference[:, 2] == 0.0).all()  # No Delta, no charge

            assert (image_difference[0:3, 3] == 0.0).all()  # No charge in front of Delta 2
            assert image_difference[3, 3] < 0.0  # Delta 2 loses charge
            assert image_difference[4, 3] > 0.0  # Delta 2 trail

            assert (image_difference[0:2, 4] == 0.0).all()  # No charge in front of Delta 3
            assert image_difference[2, 4] < 0.0  # Delta 3 loses charge
            assert (image_difference[3:5, 4] > 0.0).all()  # Delta 3 trail

    class TestRectangularImageDimensions:

        def test__horizontal_line__rectangular_image_odd_x_odd(self):
            image_pre_cti = np.zeros((5, 7))
            image_pre_cti[2, :] += 100

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0,
                                              well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=True, readout_offset=0)

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[0:2, :] == 0.0).all()
            assert (image_difference[2, :] < 0.0).all()
            assert (image_difference[3:-1, :] > 0.0).all()

        def test__horizontal_line__rectangular_image_even_x_even(self):
            
            image_pre_cti = np.zeros((4, 6))
            image_pre_cti[2, :] += 100

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0,
                                              well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=True, readout_offset=0)

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[0:2, :] == 0.0).all()
            assert (image_difference[2, :] < 0.0).all()
            assert (image_difference[3:-1, :] > 0.0).all()

        def test__horizontal_line__rectangular_image_even_x_odd(self):
            image_pre_cti = np.zeros((4, 7))
            image_pre_cti[2, :] += 100

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0,
                                              well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=True, readout_offset=0)

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[0:2, :] == 0.0).all()
            assert (image_difference[2, :] < 0.0).all()
            assert (image_difference[3:-1, :] > 0.0).all()

        def test__horizontal_line__rectangular_image_odd_x_even(self):
            image_pre_cti = np.zeros((5, 6))
            image_pre_cti[2, :] += 100

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0,
                                              well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=True, readout_offset=0)

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)
            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[0:2, :] == 0.0).all()
            assert (image_difference[2, :] < 0.0).all()
            assert (image_difference[3:-1, :] > 0.0).all()

        def test__vertical_line__rectangular_image_odd_x_odd(self):
            image_pre_cti = np.zeros((3, 5))
            image_pre_cti[:, 2] += 100

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0,
                                              well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=True, readout_offset=0)

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)
            
            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[:, 0:2] == 0.0).all()
            assert (image_difference[:, 3:-1] == 0.0).all()
            assert (image_difference[:, 2] < 0.0).all()

        def test__vertical_line__rectangular_image_even_x_even(self):
            
            image_pre_cti = np.zeros((4, 6))
            image_pre_cti[:, 2] += 100

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0,
                                              well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=True, readout_offset=0)

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[:, 0:2] == 0.0).all()
            assert (image_difference[:, 3:-1] == 0.0).all()
            assert (image_difference[:, 2] < 0.0).all()

        def test__vertical_line__rectangular_image_even_x_odd(self):
            
            image_pre_cti = np.zeros((4, 7))
            image_pre_cti[:, 2] += 100

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0,
                                              well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=True, readout_offset=0)

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[:, 0:2] == 0.0).all()
            assert (image_difference[:, 3:-1] == 0.0).all()
            assert (image_difference[:, 2] < 0.0).all()

        def test__vertical_line__rectangular_image_odd_x_even(self):
            
            image_pre_cti = np.zeros((5, 6))
            image_pre_cti[:, 2] += 100

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0,
                                              well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=True, readout_offset=0)

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[:, 0:2] == 0.0).all()
            assert (image_difference[:, 3:-1] == 0.0).all()
            assert (image_difference[:, 2] < 0.0).all()

        def test__delta_functions__add_cti_only_behind_them__odd_x_odd(self):
            
            image_pre_cti = np.zeros((5, 7))
            image_pre_cti[1, 1] += 100  # Delta 1
            image_pre_cti[3, 3] += 100  # Delta 2
            image_pre_cti[2, 4] += 100  # Delta 3

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0,
                                              well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=True, readout_offset=0)

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[:, 0] == 0.0).all()  # No Delta, no charge

            assert (image_difference[0, 1] == 0.0)  # No charge in front of Delta 1
            assert (image_difference[1, 1] < 0.0)  # Delta 1 loses charge
            assert (image_difference[2:5, 1] > 0.0).all()  # Delta 1 trails

            assert (image_difference[:, 2] == 0.0).all()  # No Delta, no charge

            assert (image_difference[0:3, 3] == 0.0).all()  # No charge in front of Delta 2
            assert image_difference[3, 3] < 0.0  # Delta 2 loses charge
            assert image_difference[4, 3] > 0.0  # Delta 2 trail

            assert (image_difference[0:2, 4] == 0.0).all()  # No charge in front of Delta 3
            assert image_difference[2, 4] < 0.0  # Delta 3 loses charge
            assert (image_difference[3:5, 4] > 0.0).all()  # Delta 3 trail

            assert (image_difference[:, 5] == 0.0).all()  # No Delta, no charge
            assert (image_difference[:, 6] == 0.0).all()  # No Delta, no charge

        def test__delta_functions__add_cti_only_behind_them__even_x_even(self):
            
            image_pre_cti = np.zeros((6, 8))
            image_pre_cti[1, 1] += 100  # Delta 1
            image_pre_cti[3, 3] += 100  # Delta 2
            image_pre_cti[2, 4] += 100  # Delta 3

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0,
                                              well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=True, readout_offset=0)

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[:, 0] == 0.0).all()  # No Delta, no charge

            assert (image_difference[0, 1] == 0.0)  # No charge in front of Delta 1
            assert (image_difference[1, 1] < 0.0)  # Delta 1 loses charge
            assert (image_difference[2:5, 1] > 0.0).all()  # Delta 1 trails

            assert (image_difference[:, 2] == 0.0).all()  # No Delta, no charge

            assert (image_difference[0:3, 3] == 0.0).all()  # No charge in front of Delta 2
            assert image_difference[3, 3] < 0.0  # Delta 2 loses charge
            assert image_difference[4, 3] > 0.0  # Delta 2 trail

            assert (image_difference[0:2, 4] == 0.0).all()  # No charge in front of Delta 3
            assert image_difference[2, 4] < 0.0  # Delta 3 loses charge
            assert (image_difference[3:5, 4] > 0.0).all()  # Delta 3 trail

            assert (image_difference[:, 5] == 0.0).all()  # No Delta, no charge
            assert (image_difference[:, 6] == 0.0).all()  # No Delta, no charge
            assert (image_difference[:, 7] == 0.0).all()  # No Delta, no charge

        def test__delta_functions__add_cti_only_behind_them__even_x_odd(self):
            
            image_pre_cti = np.zeros((6, 7))
            image_pre_cti[1, 1] += 100  # Delta 1
            image_pre_cti[3, 3] += 100  # Delta 2
            image_pre_cti[2, 4] += 100  # Delta 3

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0,
                                              well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=True, readout_offset=0)

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[:, 0] == 0.0).all()  # No Delta, no charge

            assert (image_difference[0, 1] == 0.0)  # No charge in front of Delta 1
            assert (image_difference[1, 1] < 0.0)  # Delta 1 loses charge
            assert (image_difference[2:5, 1] > 0.0).all()  # Delta 1 trails

            assert (image_difference[:, 2] == 0.0).all()  # No Delta, no charge

            assert (image_difference[0:3, 3] == 0.0).all()  # No charge in front of Delta 2
            assert image_difference[3, 3] < 0.0  # Delta 2 loses charge
            assert image_difference[4, 3] > 0.0  # Delta 2 trail

            assert (image_difference[0:2, 4] == 0.0).all()  # No charge in front of Delta 3
            assert image_difference[2, 4] < 0.0  # Delta 3 loses charge
            assert (image_difference[3:5, 4] > 0.0).all()  # Delta 3 trail

            assert (image_difference[:, 5] == 0.0).all()  # No Delta, no charge
            assert (image_difference[:, 6] == 0.0).all()  # No Delta, no charge

        def test__delta_functions__add_cti_only_behind_them__odd_x_even(self):
            image_pre_cti = np.zeros((5, 6))

            image_pre_cti[1, 1] += 100  # Delta 1
            image_pre_cti[3, 3] += 100  # Delta 2
            image_pre_cti[2, 4] += 100  # Delta 3

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0,
                                              well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=True, readout_offset=0)

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            image_difference = image_post_cti - image_pre_cti

            assert (image_difference[:, 0] == 0.0).all()  # No Delta, no charge

            assert (image_difference[0, 1] == 0.0)  # No charge in front of Delta 1
            assert (image_difference[1, 1] < 0.0)  # Delta 1 loses charge
            assert (image_difference[2:5, 1] > 0.0).all()  # Delta 1 trails

            assert (image_difference[:, 2] == 0.0).all()  # No Delta, no charge

            assert (image_difference[0:3, 3] == 0.0).all()  # No charge in front of Delta 2
            assert image_difference[3, 3] < 0.0  # Delta 2 loses charge
            assert image_difference[4, 3] > 0.0  # Delta 2 trail

            assert (image_difference[0:2, 4] == 0.0).all()  # No charge in front of Delta 3
            assert image_difference[2, 4] < 0.0  # Delta 3 loses charge
            assert (image_difference[3:5, 4] > 0.0).all()  # Delta 3 trail

            assert (image_difference[:, 5] == 0.0).all()  # No Delta, no charge

    class TestChargeInjectionMode:

        def test__charge_injection_parameters_on__makes_trails_longer(self):

            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[2, :] += 100

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0,
                                              well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                charge_injection_mode=False, readout_offset=0)


            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            settings_charge_inj = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                           charge_injection_mode=True, readout_offset=0)

            image_post_cti_charge_inj = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings_charge_inj, correct_cti=False)

            image_difference = image_post_cti_charge_inj - image_post_cti

            assert (image_difference[0:2, :] == 0.0).all()  # First four rows should all remain zero
            assert (image_difference[2,
                    :] < 0.0).all()  # With charge injection mode on, more transfers, so more charge lost
            assert (image_difference[3:-1, :] > 0.0).all()  # And therefore, trails are brighter

    class TestDensityVary:

        def test__horizontal_line__different_density_in_each_column(self):

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                charge_injection_mode=False, readout_offset=0)

            ccd = arctic_params.CCD(well_fill_beta=0.8, well_notch_depth=0.01)
            species = [arctic_params.Species(trap_density=10., trap_lifetime=1.),
                       arctic_params.Species(trap_density=20., trap_lifetime=1.),
                       arctic_params.Species(trap_density=30., trap_lifetime=1.),
                       arctic_params.Species(trap_density=40., trap_lifetime=1.),
                       arctic_params.Species(trap_density=50., trap_lifetime=1.), ]

            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[2, :] += 100

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=species, ccd=ccd, settings=settings,
                                                  correct_cti=False, use_poisson_densities=True)

            # noinspection PyUnresolvedReferences
            assert (image_post_cti[2:5, 0] != image_post_cti[2:5, 1]).all()
            # noinspection PyUnresolvedReferences
            assert (image_post_cti[2:5, 1] != image_post_cti[2:5, 2]).all()
            # noinspection PyUnresolvedReferences
            assert (image_post_cti[2:5, 2] != image_post_cti[2:5, 3]).all()
            # noinspection PyUnresolvedReferences
            assert (image_post_cti[2:5, 3] != image_post_cti[2:5, 4]).all()


class TestArcticCorrectCTI:
    
    class TestSquareDimensions:

        def test__horizontal_line__corrected_image_more_like_original(self):
            
            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[2, :] += 100
            
            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0, well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=False, readout_offset=0)

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)
            
            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = pyarctic.call_arctic(image=image_post_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=True)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__vertical_line__corrected_image_more_like_original(self):

            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[:, 2] += 100

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0, well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=False, readout_offset=0)

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = pyarctic.call_arctic(image=image_post_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=True)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__delta_functions__corrected_image_more_like_original(self):

            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[1, 1] += 100  # Delta 1
            image_pre_cti[3, 3] += 100  # Delta 2
            image_pre_cti[2, 4] += 100  # Delta 3

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0, well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=False, readout_offset=0)


            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = pyarctic.call_arctic(image=image_post_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=True)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__decrease_niter__worse_correction(self):

            image_pre_cti = np.zeros((5, 5))
            image_pre_cti[2, :] += 100

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0,
                                              well_fill_beta=0.8, well_fill_gamma=0.0)

            settings_niter_5 = arctic_settings.Settings(well_depth=84700, niter=5, express=5, n_levels=2000,
                                                charge_injection_mode=False, readout_offset=0)

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings_niter_5, correct_cti=False)

            image_correct_cti = pyarctic.call_arctic(image=image_post_cti, species=[species], ccd=ccd,
                                                  settings=settings_niter_5, correct_cti=True)

            image_difference_niter_5 = image_correct_cti - image_pre_cti

            settings_niter_3 = arctic_settings.Settings(well_depth=84700, niter=3, express=5, n_levels=2000,
                                                         charge_injection_mode=False, readout_offset=0)

            image_correct_cti = pyarctic.call_arctic(image=image_post_cti, species=[species], ccd=ccd,
                                                  settings=settings_niter_3, correct_cti=True)

            image_difference_niter_3 = image_correct_cti - image_pre_cti

            # noinspection PyUnresolvedReferences
            assert (abs(image_difference_niter_5) <= abs(
                image_difference_niter_3)).all()  # First four rows should all remain zero

    class TestRectangularImageDimensions:

        def test__horizontal_line__odd_x_odd(self):
            image_pre_cti = np.zeros((5, 3))
            image_pre_cti[2, :] += 100

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0, well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=False, readout_offset=0)


            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = pyarctic.call_arctic(image=image_post_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=True)


            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__horizontal_line__even_x_even(self):

            image_pre_cti = np.zeros((6, 4))
            image_pre_cti[2, :] += 100

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0, well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=False, readout_offset=0)

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = pyarctic.call_arctic(image=image_post_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=True)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__horizontal_line__even_x_odd(self):

            image_pre_cti = np.zeros((6, 3))
            image_pre_cti[2, :] += 100

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0, well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=False, readout_offset=0)

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = pyarctic.call_arctic(image=image_post_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=True)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__horizontal_line__odd_x_even(self):
            image_pre_cti = np.zeros((5, 4))
            image_pre_cti[2, :] += 100

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0, well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=False, readout_offset=0)

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = pyarctic.call_arctic(image=image_post_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=True)


            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__veritcal_line__odd_x_odd(self):
            image_pre_cti = np.zeros((5, 3))
            image_pre_cti[:, 2] += 100

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0, well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=False, readout_offset=0)

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = pyarctic.call_arctic(image=image_post_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=True)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__veritcal_line__even_x_even(self):
            image_pre_cti = np.zeros((6, 4))
            image_pre_cti[:, 2] += 100

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0, well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=False, readout_offset=0)

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = pyarctic.call_arctic(image=image_post_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=True)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__veritcal_line__even_x_odd(self):
            image_pre_cti = np.zeros((6, 3))
            image_pre_cti[:, 2] += 100

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0, well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=False, readout_offset=0)

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = pyarctic.call_arctic(image=image_post_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=True)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__veritcal_line__odd_x_even(self):

            image_pre_cti = np.zeros((5, 4))
            image_pre_cti[:, 2] += 100

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0, well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=False, readout_offset=0)

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = pyarctic.call_arctic(image=image_post_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=True)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__delta_functions__odd_x_odd(self):

            image_pre_cti = np.zeros((5, 7))
            image_pre_cti[1, 1] += 100  # Delta 1
            image_pre_cti[3, 3] += 100  # Delta 2
            image_pre_cti[2, 4] += 100  # Delta 3

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0, well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=False, readout_offset=0)

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = pyarctic.call_arctic(image=image_post_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=True)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__delta_functions__even_x_even(self):

            image_pre_cti = np.zeros((6, 8))
            image_pre_cti[1, 1] += 100  # Delta 1
            image_pre_cti[3, 3] += 100  # Delta 2
            image_pre_cti[2, 4] += 100  # Delta 3

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0, well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=False, readout_offset=0)

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = pyarctic.call_arctic(image=image_post_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=True)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__delta_functions__even_x_odd(self):

            image_pre_cti = np.zeros((6, 7))
            image_pre_cti[1, 1] += 100  # Delta 1
            image_pre_cti[3, 3] += 100  # Delta 2
            image_pre_cti[2, 4] += 100  # Delta 3

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0, well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=False, readout_offset=0)

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = pyarctic.call_arctic(image=image_post_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=True)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero

        def test__delta_functions__odd_x_even(self):

            image_pre_cti = np.zeros((5, 8))
            image_pre_cti[1, 1] += 100  # Delta 1
            image_pre_cti[3, 3] += 100  # Delta 2
            image_pre_cti[2, 4] += 100  # Delta 3

            species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
            ccd = arctic_params.CCD(well_notch_depth=0.01, well_fill_alpha=1.0, well_fill_beta=0.8, well_fill_gamma=0.0)

            settings = arctic_settings.Settings(well_depth=84700, niter=1, express=5, n_levels=2000,
                                                         charge_injection_mode=False, readout_offset=0)

            image_post_cti = pyarctic.call_arctic(image=image_pre_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=False)

            image_difference_1 = image_post_cti - image_pre_cti

            image_correct_cti = pyarctic.call_arctic(image=image_post_cti, species=[species], ccd=ccd,
                                                  settings=settings, correct_cti=True)

            image_difference_2 = image_correct_cti - image_pre_cti

            assert (image_difference_2 <= abs(image_difference_1)).all()  # First four rows should all remain zero


from codecs import open
from os.path import abspath, dirname, join
from subprocess import call

from setuptools import Command, find_packages, setup

from autocti import __version__

this_dir = abspath(dirname(__file__))
with open(join(this_dir, 'README.md'), encoding='utf-8') as file:
    long_description = file.read()


class RunTests(Command):
    """Run all tests."""
    description = 'run tests'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        """Run all tests!"""
        errno = call(['py.test', '--cov=autocti', '--cov-report=term-missing'])
        raise SystemExit(errno)


setup(
    name='autocti',
    version=__version__,
    description='Automated Charge Transfer Inefficiency Modeling',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Jammy2211/PyAutoCTI',
    author='James Nightingale and Richard Hayes',
    author_email='james.w.nightingale@durham.ac.uk',
    include_package_data=True,
    license='MIT License',
    classifiers=[
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Physics',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
    keywords='cli',
    packages=find_packages(exclude=['docs', 'tests*', 'workspace', 'workspace/*', 'workspace_jam', 'worspace_jam/*']),
    install_requires=['numpy',
                      'scipy',
                      'astropy',
                      'matplotlib',
                      'pymultinest',
                      'getDist',
                      'autofit==0.20.1'
                      ],
    extras_require={
        'test': ['coverage', 'pytest', 'pytest-cov'],
    },
    cmdclass={'test': RunTests},
)



from autofit import conf
from autofit.tools import path_util
from autocti.data import mask as msk
from autocti.charge_injection import ci_data
from autocti.charge_injection import ci_pattern

from autocti.model import arctic_settings

from workspace_jam.tools.data_makers import tools

from multiprocessing import Pool

import os
import sys

### NOTE - if you have not already, complete the setup in 'workspace/runners/cosma/setup' before continuing with this
### cosma pipeline script.

# Welcome to the Cosma pipeline runner. Hopefully, you're familiar with runners at this point, and have been using them
# with PyAutoCTI to model CTI on your laptop. If not, I'd recommend you get used to doing that, before trying to
# run PyAutoCTI on a super-computer. You need some familiarity with the software and before trying to model a large
# amount of charge injection imaging on a supercomputer!

# If you are ready, then let me take you through the Cosma runner. It is remarkably similar to the ordinary pipeline
# runners you're used to, however it makes a few changes for running jobs on cosma:

# 1) The data path is over-written to the path '/cosma5/data/durham/cosma_username/autocti/data' as opposed to the
#    workspace. As we saw in the setup, on cosma we don't store our data in our workspace.

# 2) The output path is over-written to the path '/cosma5/data/durham/cosma_username/autocti/output' as opposed to
#    the workspace. This is for the same reason as the data.

# Given your username is where your data is stored, you'll need to put your cosma username here.
cosma_username = 'dc-nigh1'

# The cosma path where the data and output are stored.
cosma_path = '/cosma7/data/dp004/dc-nigh1/autocti/'

# Get the relative path to the config files and output folder in our workspace.
workspace_path = '{}/../../../../'.format(os.path.dirname(os.path.realpath(__file__)))

# Use this path to explicitly set the config path, and override the output path with the Cosma path.
conf.instance = conf.Config(config_path=workspace_path + 'config', output_path=cosma_path + 'output')

# Lets take a look at a Cosma batch script, which can be found at 'workspace/runners/cosma/batch/pipeline_runner_cosma'.
# When we submit a PyAutoLens job to Cosma, we submit a 'batch' of jobs, whereby each job will run on one CPU of Cosma.
# Thus, if our lens sample contains, lets say, 4 lenses, we'd submit 4 jobs at the same time where each job applies
# our pipeline to each image.

# The fifth line of this batch script - '#SBATCH --array=1-4' is what species this. Its telling Cosma we're going to
# run 4 jobs, and the id's of those jobs will be numbered from 1 to 4. Infact, these ids are passed to this runner,
# and we'll use them to ensure that each jobs loads a different image. Lets get the cosma array id for our job.
batch_id = int(sys.argv[1])

# For a given COSMA run, we will assume all of the charge injection data-sets have identical properties. That is, they
# all have the same dimensions (shape), charge injection regions (ci_regions), charge injection normalizations
# (normalizations), charge injection patterns (patterns) and frame geometries defining the direction clocking and CTI
# (frame_geometry).

# If you need to change these value for a COSMA run, I recommend making a different runner script.

ci_data_type = 'ci_images_uniform' # Charge injection data consisting of 2 images with uniform injections.
ci_data_model = 'serial_x2'

ci_data_resolution = []
ci_data_resolution.append('')
ci_data_resolution.append('low_resolution') # Index 1
ci_data_resolution.append('mid_resolution') # Index 2
ci_data_resolution.append('high_resolution') # Index 3

ci_data_resolution = ci_data_resolution[batch_id]

# Now, I just want to really drive home what the above line is doing. For every job we run on Cosma, the cosma_array_id
# will be different. That is, job 1 will get a cosma_array_id of 1, job 2 will get an id of 2, and so on. This is our
# only unique identifier of every job, thus its our only hope of specifying for each job which image they load!

# Fortunately, we're used to specifying the data name as a string, so that our pipeline can be applied to multiple
# images with ease. On Cosma, we can apply the same logic, but put these strings in a list such that each Cosma job
# loads a different lens name based on its ID. neat, huh?

# Create the path where the data will be loaded from, which in this case is
# '/workspace/data/ci_images_uniform/serial_x2_species/high_res/'
ci_data_path = path_util.make_and_return_path_from_path_and_folder_names(
    path=cosma_path, folder_names=['data', ci_data_type, ci_data_model, ci_data_resolution])

# This tools loads the correct shape of the image and frame geometry, for the ci_data_resoultion string.
shape, frame_geometry = \
    tools.serial_shape_and_frame_geometry_from_ci_data_resolution(ci_data_resolution=ci_data_resolution)

# The charge injection regions on the CCD, which in this case is 7 equally spaced rectangular blocks.
ci_regions = [(0, 30, 51, shape[1]-20), (330, 360, 51, shape[1]-20),
              (660, 690, 51, shape[1]-20), (990, 1020, 51, shape[1]-20),
              (1320, 1350, 51, shape[1]-20), (1650, 1680, 51, shape[1]-20),
              (1980, 2010, 51, shape[1]-20)]

# The normalization of the charge injection for each image.
normalizations=[100.0, 500.0, 1000.0, 5000.0, 10000.0, 25000.0, 50000.0, 84700.0]

# Create the charge injection pattern objects used for this pipeline.
patterns = ci_pattern.uniform_from_lists(normalizations=normalizations, regions=ci_regions)

ci_datas = []

for index, normalization in enumerate(normalizations):

    ci_datas.append(ci_data.ci_data_from_fits(
                 frame_geometry=frame_geometry, ci_pattern=patterns[index],
                 image_path=ci_data_path + 'image_' + str(int(normalization)) + '.fits',
                 ci_pre_cti_path=ci_data_path +'ci_pre_cti_' + str(int(normalization)) + '.fits',
                 noise_map_from_single_value=4.0))

serial_cti_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=2, n_levels=2000,
                                               charge_injection_mode=False, readout_offset=0)
cti_settings = arctic_settings.ArcticSettings(serial=serial_cti_settings)

# Running a pipeline is exactly the same as we're used to. We import it, make it, and run it, noting that we can
# use the lens_name's to ensure each job outputs its results to a different directory.

from workspace_jam.pipelines.initialize import serial_x2

pipeline_initialize = serial_x2.make_pipeline(phase_folders=[ci_data_type, ci_data_model, ci_data_resolution])

from workspace_jam.pipelines.normal.from_initialize import serial_x2

pipeline_normal = serial_x2.make_pipeline(phase_folders=[ci_data_type, ci_data_model, ci_data_resolution])

pipeline = pipeline_initialize + pipeline_normal

pipeline.run(ci_datas=ci_datas, cti_settings=cti_settings, pool=Pool(processes=8), assert_optimizer_pickle_matches=False)


from autofit.tools import path_util
from autocti.data import cti_image
from autocti.charge_injection import ci_frame

from autocti.model import arctic_settings
from autocti.model import arctic_params

import os
import numpy as np

# This tool allows one to make simulated images of parallel charge transfer inefficiency, typically using simple image
# configurations to highlight the effects of capture and trailing.

# The 'data name' is the name of the lens in the data folder, e.g:

# The image will be output as '/workspace/data/data_name/image.fits'.
# The pre cti image will be output as '/workspace/data/data_name/image_pre_cti.fits'.

# Now, lets output this simulated ccd-data to the data folder. First, we'll get a relative path.
workspace_path = '{}/../'.format(os.path.dirname(os.path.realpath(__file__)))

data_type = 'parallel_image_horizontal_line'

# Setup the path to the data, where it will ultimately be output.
data_path = path_util.make_and_return_path_from_path_and_folder_names(path=workspace_path,
                                                                      folder_names=['data', data_type])

shape = (5, 5) # The shape determines the shape of the image simulated.

# Lets use this shape to create a NumPy array of zeros of this shape, and add a horiztonal line acrosss the image
# perpedicular to the direction of read-out and thus CTI.
image = np.zeros(shape=shape)
image[1,0:5] = 1000.0

# The frame geometry of the image being simuated.
frame_geometry = ci_frame.QuadGeometryEuclid.bottom_left()
# frame_geometry = ci_frame.QuadGeometryEuclid.bottom_right()
# frame_geometry = ci_frame.QuadGeometryEuclid.top_left()
# frame_geometry = ci_frame.QuadGeometryEuclid.top_right()


# Set this image up as a CTI image, so that we can add CTI to it.
image_pre_cti = cti_image.ImageFrame(frame_geometry=frame_geometry, array=image)

# The CTI settings of arCTIc, which models the CCD read-out including CTI.
parallel_cti_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=2, n_levels=2000,
                                                 charge_injection_mode=False, readout_offset=0)
cti_settings = arctic_settings.ArcticSettings(parallel=parallel_cti_settings)

# The CTI model parameters of arCTIc, which includes each trap species density / lifetime and the CCD properties for
# parallel charge transfer.
parallel_species = arctic_params.Species(trap_density=0.5, trap_lifetime=4.0)
parallel_ccd = arctic_params.CCD(well_notch_depth=1.0e-4, well_fill_beta=0.58, well_fill_gamma=0.0, well_fill_alpha=1.0)
cti_params = arctic_params.ArcticParams(parallel_species=[parallel_species], parallel_ccd=parallel_ccd)

# Use the pre-cti image to create a post-cti image, by passing it the arctic setting and CTI model above.
image_post_cti = image_pre_cti.add_cti_to_image(cti_params=cti_params, cti_settings=cti_settings)

# Setup the post cti image as an image frame (this should be inbuilt into array_finalize functions in the future)
image_post_cti = cti_image.ImageFrame(frame_geometry=frame_geometry, array=image_post_cti)

# Now, output every image to the data folder.
image_pre_cti.output_as_fits(file_path=data_path + '/image_pre_cti.fits', overwrite=True)
image_post_cti.output_as_fits(file_path=data_path + '/image_post_cti.fits', overwrite=True)


from os import path

import numpy as np
import pytest
from autofit import conf
from autofit.mapper import model_mapper as mm
from autofit.mapper import prior_model
from autofit.optimize import non_linear as nl
from autofit.tools.pipeline import ResultsCollection

from autocti import exc
from autocti.charge_injection import ci_data as data
from autocti.charge_injection import ci_fit
from autocti.charge_injection import ci_frame
from autocti.data import mask as msk
from autocti.model import arctic_params
from autocti.model import arctic_settings
from autocti.pipeline import phase as ph

pytestmark = pytest.mark.filterwarnings(
    "ignore:Using a non-tuple sequence for multidimensional indexing is deprecated; use `arr[tuple(seq)]` instead of "
    "`arr[seq]`. In the future this will be interpreted as an array index, `arr[np.array(seq)]`, which will result "
    "either in an error or a different result.")

directory = path.dirname(path.realpath(__file__))


@pytest.fixture(scope="session", autouse=True)
def do_something():
    conf.instance = conf.Config('{}/../test_files/configs/phase'.format(directory))


class MockResults(object):
    def __init__(self, ci_post_ctis):
        self.ci_post_ctis = ci_post_ctis
        self.constant = mm.ModelInstance()
        self.variable = mm.ModelMapper()


class NLO(nl.NonLinearOptimizer):

    def fit(self, analysis):
        class Fitness(object):

            def __init__(self, instance_from_physical_vector, constant):
                self.result = None
                self.instance_from_physical_vector = instance_from_physical_vector
                self.constant = constant

            def __call__(self, vector):
                instance = self.instance_from_physical_vector(vector)
                for key, value in self.constant.__dict__.items():
                    setattr(instance, key, value)

                likelihood = analysis.fit(instance)
                self.result = nl.Result(instance, likelihood)

                # Return Chi squared
                return -2 * likelihood

        fitness_function = Fitness(self.variable.instance_from_physical_vector, self.constant)
        fitness_function(self.variable.prior_count * [0.5])

        return fitness_function.result


class MockPattern(object):

    def __init__(self):
        self.regions = [ci_frame.Region(region=[0, 1, 0, 1])]


@pytest.fixture(name="phase")
def make_phase():
    return ph.ParallelPhase(phase_name='test_phase', optimizer_class=NLO)


@pytest.fixture(name="cti_settings")
def make_cti_settings():
    parallel_settings = arctic_settings.Settings(well_depth=0, niter=1, express=1, n_levels=2000)
    return arctic_settings.ArcticSettings(neomode='NEO', parallel=parallel_settings)


@pytest.fixture(name="frame_geometry")
def make_frame_geometry():
    return ci_frame.QuadGeometryEuclid.bottom_left()


@pytest.fixture(name="ci_pattern")
def make_ci_pattern():
    return MockPattern()


@pytest.fixture(name="ci_data")
def make_ci_data(frame_geometry, ci_pattern):
    image = np.ones((3, 3))
    noise = np.ones((3, 3))
    ci_pre_cti = np.ones((3, 3))
    frame = ci_frame.ChInj(frame_geometry=frame_geometry, ci_pattern=ci_pattern)
    return data.CIData(image=image, noise_map=noise, ci_pre_cti=ci_pre_cti, ci_pattern=ci_pattern,
                       ci_frame=frame)


@pytest.fixture(name="results")
def make_results():
    return MockResults(ci_post_ctis=[np.ones((10, 10)), np.ones((10, 10))])


@pytest.fixture(name="results_collection")
def make_results_collection():
    return ResultsCollection()


class TestPhase(object):

    def test_param_names(self, phase):
        phase.parallel_species = [prior_model.PriorModel(arctic_params.Species),
                                  prior_model.PriorModel(arctic_params.Species)]

        assert phase.optimizer.variable.param_names == ['parallel_species_0_trap_density',
                                                        'parallel_species_0_trap_lifetime',
                                                        'parallel_species_1_trap_density',
                                                        'parallel_species_1_trap_lifetime']

        phase = ph.ParallelPhase(phase_name='test_phase',
                                 parallel_species=[prior_model.PriorModel(arctic_params.Species),
                                                   prior_model.PriorModel(arctic_params.Species)],
                                 optimizer_class=NLO)

        assert phase.optimizer.variable.param_names == ['parallel_species_0_trap_density',
                                                        'parallel_species_0_trap_lifetime',
                                                        'parallel_species_1_trap_density',
                                                        'parallel_species_1_trap_lifetime']

    def test__make_analysis(self, phase, ci_data, cti_settings):
        analysis = phase.make_analysis(ci_datas=[ci_data], cti_settings=cti_settings)
        assert analysis.last_results is None
        assert (analysis.ci_datas_extracted[0].image == ci_data.image).all()
        assert (analysis.ci_datas_extracted[0].noise_map == ci_data.noise_map).all()
        assert (analysis.ci_datas_full[0].image == ci_data.image).all()
        assert (analysis.ci_datas_full[0].noise_map == ci_data.noise_map).all()
        assert analysis.cti_settings == cti_settings

    def test__make_analysis__uses_mask_function(self, phase, ci_data, cti_settings):

        def mask_function(shape, ci_frame):
            return np.full(shape=shape, fill_value=False)

        phase.mask_function = mask_function

        analysis = phase.make_analysis(ci_datas=[ci_data], cti_settings=cti_settings)

        assert (analysis.ci_datas_full[0].mask == np.full(shape=(3, 3), fill_value=False)).all()

        def mask_function(shape, ci_frame):
            return np.array([[False, False, False],
                             [False, True, False],
                             [False, False, False]])

        phase.mask_function = mask_function

        analysis = phase.make_analysis(ci_datas=[ci_data], cti_settings=cti_settings)

        assert (analysis.ci_datas_full[0].mask == np.array([[False, False, False],
                                                            [False, True, False],
                                                            [False, False, False]])).all()

    def test__make_analysis__default_mask_all_empty__cosmic_ray_image_masks_mask(self, phase, ci_data, cti_settings):

        phase.cosmic_ray_parallel_buffer = 0
        phase.cosmic_ray_serial_buffer = 0
        phase.cosmic_ray_diagonal_buffer = 0

        ci_data.cosmic_ray_image = np.array([[0.0, 0.0, 0.0],
                                             [0.0, 1.0, 0.0],
                                             [0.0, 0.0, 0.0]])

        analysis = phase.make_analysis(ci_datas=[ci_data], cti_settings=cti_settings)

        assert (analysis.ci_datas_full[0].mask == np.array([[False, False, False],
                                                            [False, True, False],
                                                            [False, False, False]])).all()

        phase.cosmic_ray_parallel_buffer = 1
        phase.cosmic_ray_serial_buffer = 1
        phase.cosmic_ray_diagonal_buffer = 1

        ci_data.cosmic_ray_image = np.array([[0.0, 0.0, 0.0],
                                             [0.0, 1.0, 0.0],
                                             [0.0, 0.0, 0.0]])

        analysis = phase.make_analysis(ci_datas=[ci_data], cti_settings=cti_settings)

        assert (analysis.ci_datas_full[0].mask == np.array([[False, False, False],
                                                            [False, True, True],
                                                            [False, True, True]])).all()

        phase.cosmic_ray_parallel_buffer = 2
        phase.cosmic_ray_serial_buffer = 1
        phase.cosmic_ray_diagonal_buffer = 1

        ci_data.cosmic_ray_image = np.array([[1.0, 0.0, 0.0],
                                             [0.0, 0.0, 0.0],
                                             [0.0, 0.0, 0.0]])

        analysis = phase.make_analysis(ci_datas=[ci_data], cti_settings=cti_settings)

        assert (analysis.ci_datas_full[0].mask == np.array([[True, True, False],
                                                            [True, True, False],
                                                            [True, False, False]])).all()

    def test__make_analysis__uses_parallel_and_serial_front_edge_or_trials_for_mask(self, phase, ci_data, cti_settings):

        phase.parallel_front_edge_mask_rows = (0, 1)
        phase.parallel_trails_mask_rows = None
        phase.serial_front_edge_mask_columns = None
        phase.serial_trails_mask_columns = None

        analysis = phase.make_analysis(ci_datas=[ci_data], cti_settings=cti_settings)

        assert (analysis.ci_datas_full[0].mask == np.array([[True, False, False],
                                                            [False, False, False],
                                                            [False, False, False]])).all()

        phase.parallel_front_edge_mask_rows = None
        phase.parallel_trails_mask_rows = (0, 2)
        phase.serial_front_edge_mask_columns = None
        phase.serial_trails_mask_columns = None

        analysis = phase.make_analysis(ci_datas=[ci_data], cti_settings=cti_settings)

        assert (analysis.ci_datas_full[0].mask == np.array([[False, False, False],
                                                            [True, False, False],
                                                            [True, False, False]])).all()


        phase.parallel_front_edge_mask_rows = None
        phase.parallel_trails_mask_rows = None
        phase.serial_front_edge_mask_columns = (0, 1)
        phase.serial_trails_mask_columns = None

        analysis = phase.make_analysis(ci_datas=[ci_data], cti_settings=cti_settings)

        assert (analysis.ci_datas_full[0].mask == np.array([[True, False, False],
                                                            [False, False, False],
                                                            [False, False, False]])).all()

        phase.parallel_front_edge_mask_rows = None
        phase.parallel_trails_mask_rows = None
        phase.serial_front_edge_mask_columns = None
        phase.serial_trails_mask_columns = (0, 2)

        analysis = phase.make_analysis(ci_datas=[ci_data], cti_settings=cti_settings)

        assert (analysis.ci_datas_full[0].mask == np.array([[False, True, True],
                                                            [False, False, False],
                                                            [False, False, False]])).all()

        phase.parallel_front_edge_mask_rows = (0, 1)
        phase.parallel_trails_mask_rows = (1, 2)
        phase.serial_front_edge_mask_columns = (0, 1)
        phase.serial_trails_mask_columns = (0, 2)

        analysis = phase.make_analysis(ci_datas=[ci_data], cti_settings=cti_settings)

        assert (analysis.ci_datas_full[0].mask == np.array([[True,  True,  True],
                                                            [False, False, False],
                                                            [True, False, False]])).all()

    def test__parallel_phase__if_trap_lifetime_not_ascending__raises_exception(self, phase, ci_data, cti_settings):
        analysis = phase.make_analysis(ci_datas=[ci_data], cti_settings=cti_settings)

        species_0 = arctic_params.Species(trap_density=0.0, trap_lifetime=1.0)
        species_1 = arctic_params.Species(trap_density=0.0, trap_lifetime=2.0)
        cti_params = arctic_params.ArcticParams(parallel_species=[species_0, species_1])

        analysis.check_trap_lifetimes_are_ascending(cti_params=cti_params)

        species_0 = arctic_params.Species(trap_density=0.0, trap_lifetime=1.0)
        species_1 = arctic_params.Species(trap_density=0.0, trap_lifetime=2.0)
        species_2 = arctic_params.Species(trap_density=0.0, trap_lifetime=3.0)
        cti_params = arctic_params.ArcticParams(parallel_species=[species_0, species_1, species_2])

        analysis.check_trap_lifetimes_are_ascending(cti_params=cti_params)

        species_0 = arctic_params.Species(trap_density=0.0, trap_lifetime=2.0)
        species_1 = arctic_params.Species(trap_density=0.0, trap_lifetime=1.0)
        cti_params = arctic_params.ArcticParams(parallel_species=[species_0, species_1])

        with pytest.raises(exc.PriorException):
            analysis.check_trap_lifetimes_are_ascending(cti_params=cti_params)

        species_0 = arctic_params.Species(trap_density=0.0, trap_lifetime=1.0)
        species_1 = arctic_params.Species(trap_density=0.0, trap_lifetime=2.0)
        species_2 = arctic_params.Species(trap_density=0.0, trap_lifetime=1.0)
        cti_params = arctic_params.ArcticParams(parallel_species=[species_0, species_1, species_2])

        with pytest.raises(exc.PriorException):
            analysis.check_trap_lifetimes_are_ascending(cti_params=cti_params)

        species_0 = arctic_params.Species(trap_density=0.0, trap_lifetime=4.0)
        species_1 = arctic_params.Species(trap_density=0.0, trap_lifetime=2.0)
        species_2 = arctic_params.Species(trap_density=0.0, trap_lifetime=3.0)
        cti_params = arctic_params.ArcticParams(parallel_species=[species_0, species_1, species_2])

        with pytest.raises(exc.PriorException):
            analysis.check_trap_lifetimes_are_ascending(cti_params=cti_params)

        species_0 = arctic_params.Species(trap_density=0.0, trap_lifetime=5.0)
        species_1 = arctic_params.Species(trap_density=0.0, trap_lifetime=4.0)
        species_2 = arctic_params.Species(trap_density=0.0, trap_lifetime=3.0)
        cti_params = arctic_params.ArcticParams(parallel_species=[species_0, species_1, species_2])

        with pytest.raises(exc.PriorException):
            analysis.check_trap_lifetimes_are_ascending(cti_params=cti_params)

    def test__serial_phase__if_trap_lifetime_not_ascending__raises_exception(self, ci_data, cti_settings):
        phase = ph.SerialPhase(phase_name='test_phase', optimizer_class=NLO)

        analysis = phase.make_analysis(ci_datas=[ci_data], cti_settings=cti_settings)

        species_0 = arctic_params.Species(trap_density=0.0, trap_lifetime=1.0)
        species_1 = arctic_params.Species(trap_density=0.0, trap_lifetime=2.0)
        cti_params = arctic_params.ArcticParams(serial_species=[species_0, species_1])

        analysis.check_trap_lifetimes_are_ascending(cti_params=cti_params)

        species_0 = arctic_params.Species(trap_density=0.0, trap_lifetime=1.0)
        species_1 = arctic_params.Species(trap_density=0.0, trap_lifetime=2.0)
        species_2 = arctic_params.Species(trap_density=0.0, trap_lifetime=3.0)
        cti_params = arctic_params.ArcticParams(serial_species=[species_0, species_1, species_2])

        analysis.check_trap_lifetimes_are_ascending(cti_params=cti_params)

        species_0 = arctic_params.Species(trap_density=0.0, trap_lifetime=2.0)
        species_1 = arctic_params.Species(trap_density=0.0, trap_lifetime=1.0)
        cti_params = arctic_params.ArcticParams(serial_species=[species_0, species_1])

        with pytest.raises(exc.PriorException):
            analysis.check_trap_lifetimes_are_ascending(cti_params=cti_params)

        species_0 = arctic_params.Species(trap_density=0.0, trap_lifetime=1.0)
        species_1 = arctic_params.Species(trap_density=0.0, trap_lifetime=2.0)
        species_2 = arctic_params.Species(trap_density=0.0, trap_lifetime=1.0)
        cti_params = arctic_params.ArcticParams(serial_species=[species_0, species_1, species_2])

        with pytest.raises(exc.PriorException):
            analysis.check_trap_lifetimes_are_ascending(cti_params=cti_params)

        species_0 = arctic_params.Species(trap_density=0.0, trap_lifetime=4.0)
        species_1 = arctic_params.Species(trap_density=0.0, trap_lifetime=2.0)
        species_2 = arctic_params.Species(trap_density=0.0, trap_lifetime=3.0)
        cti_params = arctic_params.ArcticParams(serial_species=[species_0, species_1, species_2])

        with pytest.raises(exc.PriorException):
            analysis.check_trap_lifetimes_are_ascending(cti_params=cti_params)

        species_0 = arctic_params.Species(trap_density=0.0, trap_lifetime=5.0)
        species_1 = arctic_params.Species(trap_density=0.0, trap_lifetime=4.0)
        species_2 = arctic_params.Species(trap_density=0.0, trap_lifetime=3.0)
        cti_params = arctic_params.ArcticParams(serial_species=[species_0, species_1, species_2])

        with pytest.raises(exc.PriorException):
            analysis.check_trap_lifetimes_are_ascending(cti_params=cti_params)

    def test__parallel_serial_phase__if_trap_lifetime_not_ascending__raises_exception(self, ci_data, cti_settings):
        phase = ph.ParallelSerialPhase(phase_name='test_phase', optimizer_class=NLO)

        analysis = phase.make_analysis(ci_datas=[ci_data], cti_settings=cti_settings)

        parallel_species_0 = arctic_params.Species(trap_density=0.0, trap_lifetime=1.0)
        parallel_species_1 = arctic_params.Species(trap_density=0.0, trap_lifetime=2.0)
        serial_species_0 = arctic_params.Species(trap_density=0.0, trap_lifetime=1.0)
        serial_species_1 = arctic_params.Species(trap_density=0.0, trap_lifetime=2.0)
        cti_params = arctic_params.ArcticParams(parallel_species=[parallel_species_0, parallel_species_1],
                                                serial_species=[serial_species_0, serial_species_1])

        analysis.check_trap_lifetimes_are_ascending(cti_params=cti_params)

        parallel_species_0 = arctic_params.Species(trap_density=0.0, trap_lifetime=1.0)
        parallel_species_1 = arctic_params.Species(trap_density=0.0, trap_lifetime=2.0)
        parallel_species_2 = arctic_params.Species(trap_density=0.0, trap_lifetime=3.0)
        serial_species_0 = arctic_params.Species(trap_density=0.0, trap_lifetime=1.0)
        serial_species_1 = arctic_params.Species(trap_density=0.0, trap_lifetime=2.0)
        serial_species_2 = arctic_params.Species(trap_density=0.0, trap_lifetime=3.0)
        cti_params = arctic_params.ArcticParams(
            parallel_species=[parallel_species_0, parallel_species_1, parallel_species_2],
            serial_species=[serial_species_0, serial_species_1, serial_species_2])

        analysis.check_trap_lifetimes_are_ascending(cti_params=cti_params)

        parallel_species_0 = arctic_params.Species(trap_density=0.0, trap_lifetime=2.0)
        parallel_species_1 = arctic_params.Species(trap_density=0.0, trap_lifetime=1.0)
        serial_species_0 = arctic_params.Species(trap_density=0.0, trap_lifetime=1.0)
        serial_species_1 = arctic_params.Species(trap_density=0.0, trap_lifetime=2.0)
        cti_params = arctic_params.ArcticParams(parallel_species=[parallel_species_0, parallel_species_1],
                                                serial_species=[serial_species_0, serial_species_1])

        with pytest.raises(exc.PriorException):
            analysis.check_trap_lifetimes_are_ascending(cti_params=cti_params)

        parallel_species_0 = arctic_params.Species(trap_density=0.0, trap_lifetime=1.0)
        parallel_species_1 = arctic_params.Species(trap_density=0.0, trap_lifetime=2.0)
        serial_species_0 = arctic_params.Species(trap_density=0.0, trap_lifetime=2.0)
        serial_species_1 = arctic_params.Species(trap_density=0.0, trap_lifetime=1.0)
        cti_params = arctic_params.ArcticParams(parallel_species=[parallel_species_0, parallel_species_1],
                                                serial_species=[serial_species_0, serial_species_1])
        with pytest.raises(exc.PriorException):
            analysis.check_trap_lifetimes_are_ascending(cti_params=cti_params)

    def test__parallel_phase__total_density_within_values__if_not_true_raises_exception(self, phase, ci_data, cti_settings):

        phase.parallel_total_density_range = (1.0, 2.0)

        analysis = phase.make_analysis(ci_datas=[ci_data], cti_settings=cti_settings)

        species_0 = arctic_params.Species(trap_density=0.75, trap_lifetime=1.0)
        species_1 = arctic_params.Species(trap_density=0.75, trap_lifetime=2.0)
        cti_params = arctic_params.ArcticParams(parallel_species=[species_0, species_1])

        analysis.check_total_density_within_range(cti_params=cti_params)

        species_0 = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
        species_1 = arctic_params.Species(trap_density=0.1, trap_lifetime=2.0)
        cti_params = arctic_params.ArcticParams(parallel_species=[species_0, species_1])

        with pytest.raises(exc.PriorException):
            analysis.check_total_density_within_range(cti_params=cti_params)

        species_0 = arctic_params.Species(trap_density=1.5, trap_lifetime=1.0)
        species_1 = arctic_params.Species(trap_density=1.5, trap_lifetime=2.0)
        cti_params = arctic_params.ArcticParams(parallel_species=[species_0, species_1])

        with pytest.raises(exc.PriorException):
            analysis.check_total_density_within_range(cti_params=cti_params)

        phase.parallel_total_density_range = (10.0, 15.0)

        analysis = phase.make_analysis(ci_datas=[ci_data], cti_settings=cti_settings)

        species_0 = arctic_params.Species(trap_density=12.0, trap_lifetime=1.0)
        species_1 = arctic_params.Species(trap_density=2.0, trap_lifetime=2.0)
        cti_params = arctic_params.ArcticParams(parallel_species=[species_0, species_1])

        analysis.check_total_density_within_range(cti_params=cti_params)

        species_0 = arctic_params.Species(trap_density=9.0, trap_lifetime=1.0)
        species_1 = arctic_params.Species(trap_density=0.9, trap_lifetime=2.0)
        cti_params = arctic_params.ArcticParams(parallel_species=[species_0, species_1])

        with pytest.raises(exc.PriorException):
            analysis.check_total_density_within_range(cti_params=cti_params)

        species_0 = arctic_params.Species(trap_density=14.0, trap_lifetime=1.0)
        species_1 = arctic_params.Species(trap_density=2.0, trap_lifetime=2.0)
        cti_params = arctic_params.ArcticParams(parallel_species=[species_0, species_1])

        with pytest.raises(exc.PriorException):
            analysis.check_total_density_within_range(cti_params=cti_params)

    def test__serial_phase__total_density_within_values__if_not_true_raises_exception(self, phase, ci_data, cti_settings):

        phase = ph.SerialPhase(phase_name='test_phase', optimizer_class=NLO)

        phase.serial_total_density_range = (1.0, 2.0)

        analysis = phase.make_analysis(ci_datas=[ci_data], cti_settings=cti_settings)

        species_0 = arctic_params.Species(trap_density=0.75, trap_lifetime=1.0)
        species_1 = arctic_params.Species(trap_density=0.75, trap_lifetime=2.0)
        cti_params = arctic_params.ArcticParams(serial_species=[species_0, species_1])

        analysis.check_total_density_within_range(cti_params=cti_params)

        species_0 = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
        species_1 = arctic_params.Species(trap_density=0.1, trap_lifetime=2.0)
        cti_params = arctic_params.ArcticParams(serial_species=[species_0, species_1])

        with pytest.raises(exc.PriorException):
            analysis.check_total_density_within_range(cti_params=cti_params)

        species_0 = arctic_params.Species(trap_density=1.5, trap_lifetime=1.0)
        species_1 = arctic_params.Species(trap_density=1.5, trap_lifetime=2.0)
        cti_params = arctic_params.ArcticParams(serial_species=[species_0, species_1])

        with pytest.raises(exc.PriorException):
            analysis.check_total_density_within_range(cti_params=cti_params)

        phase.serial_total_density_range = (10.0, 15.0)

        analysis = phase.make_analysis(ci_datas=[ci_data], cti_settings=cti_settings)

        species_0 = arctic_params.Species(trap_density=12.0, trap_lifetime=1.0)
        species_1 = arctic_params.Species(trap_density=2.0, trap_lifetime=2.0)
        cti_params = arctic_params.ArcticParams(serial_species=[species_0, species_1])

        analysis.check_total_density_within_range(cti_params=cti_params)

        species_0 = arctic_params.Species(trap_density=9.0, trap_lifetime=1.0)
        species_1 = arctic_params.Species(trap_density=0.9, trap_lifetime=2.0)
        cti_params = arctic_params.ArcticParams(serial_species=[species_0, species_1])

        with pytest.raises(exc.PriorException):
            analysis.check_total_density_within_range(cti_params=cti_params)

        species_0 = arctic_params.Species(trap_density=14.0, trap_lifetime=1.0)
        species_1 = arctic_params.Species(trap_density=2.0, trap_lifetime=2.0)
        cti_params = arctic_params.ArcticParams(serial_species=[species_0, species_1])

        with pytest.raises(exc.PriorException):
            analysis.check_total_density_within_range(cti_params=cti_params)

    def test__parallel_serial_phase__total_density_within_values__if_not_true_raises_exception(self, phase, ci_data,
                                                                                               cti_settings):

        phase = ph.ParallelSerialPhase(phase_name='test_phase', optimizer_class=NLO)

        phase.parallel_total_density_range = (1.0, 2.0)
        phase.serial_total_density_range = (5.0, 6.0)

        analysis = phase.make_analysis(ci_datas=[ci_data], cti_settings=cti_settings)

        species_0 = arctic_params.Species(trap_density=0.75, trap_lifetime=1.0)
        species_1 = arctic_params.Species(trap_density=0.75, trap_lifetime=2.0)
        species_2 = arctic_params.Species(trap_density=5.5, trap_lifetime=1.0)
        species_3 = arctic_params.Species(trap_density=0.1, trap_lifetime=2.0)

        cti_params = arctic_params.ArcticParams(parallel_species=[species_0, species_1],
                                                serial_species=[species_2, species_3])

        analysis.check_total_density_within_range(cti_params=cti_params)

        phase.parallel_total_density_range = (1.0, 2.0)
        phase.serial_total_density_range = (5.0, 6.0)

        analysis = phase.make_analysis(ci_datas=[ci_data], cti_settings=cti_settings)

        species_0 = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
        species_1 = arctic_params.Species(trap_density=0.1, trap_lifetime=2.0)
        species_2 = arctic_params.Species(trap_density=5.5, trap_lifetime=1.0)
        species_3 = arctic_params.Species(trap_density=0.1, trap_lifetime=2.0)

        cti_params = arctic_params.ArcticParams(parallel_species=[species_0, species_1],
                                                serial_species=[species_2, species_3])

        with pytest.raises(exc.PriorException):
            analysis.check_total_density_within_range(cti_params=cti_params)

        phase.parallel_total_density_range = (1.0, 2.0)
        phase.serial_total_density_range = (5.0, 6.0)

        analysis = phase.make_analysis(ci_datas=[ci_data], cti_settings=cti_settings)

        species_0 = arctic_params.Species(trap_density=1.0, trap_lifetime=1.0)
        species_1 = arctic_params.Species(trap_density=0.5, trap_lifetime=2.0)
        species_2 = arctic_params.Species(trap_density=6.0, trap_lifetime=1.0)
        species_3 = arctic_params.Species(trap_density=0.1, trap_lifetime=2.0)

        cti_params = arctic_params.ArcticParams(parallel_species=[species_0, species_1],
                                                serial_species=[species_2, species_3])

        with pytest.raises(exc.PriorException):
            analysis.check_total_density_within_range(cti_params=cti_params)


    def test__customize_constant(self, results, ci_data, cti_settings):
        class MyPhase(ph.ParallelPhase):
            def pass_priors(self, previous_results):
                self.parallel_species = previous_results.last.constant.parallel_species

        parallel = arctic_params.Species()

        setattr(results.constant, "parallel_species", [parallel])

        results_collection = ResultsCollection()
        results_collection.add("first_phase", results)

        phase = MyPhase(phase_name='test_phase', optimizer_class=NLO, parallel_species=[parallel])
        phase.make_analysis([ci_data], cti_settings, results=results_collection)

        assert phase.parallel_species == [parallel]

    # noinspection PyUnresolvedReferences
    def test__default_data_extractor(self, ci_data, phase):
        ci_datas_fit = phase.extract_ci_data(ci_data, msk.Mask.empty_for_shape(ci_data.image.shape))

        assert isinstance(ci_datas_fit, data.MaskedCIData)
        assert (ci_data.image == ci_datas_fit.image).all()
        assert (ci_data.noise_map == ci_datas_fit.noise_map).all()
        assert (ci_data.ci_pre_cti == ci_datas_fit.ci_pre_cti).all()

    def test__duplication(self):
        parallel = arctic_params.Species()

        phase = ph.ParallelPhase(phase_name='test_phase', parallel_species=[parallel])

        ph.ParallelPhase(phase_name='test_phase', parallel_species=[])

        assert phase.parallel_species is not None

    def test__cti_params_for_instance(self):
        instance = mm.ModelInstance()
        instance.parallel_species = [arctic_params.Species(trap_density=0.0)]
        cti_params = ph.cti_params_for_instance(instance)

        assert cti_params.parallel_species == instance.parallel_species


class MockResult:
    noise_scaling_maps_of_ci_regions = 1
    noise_scaling_maps_of_parallel_trails = 2
    noise_scaling_maps_of_serial_trails = 3
    noise_scaling_maps_of_serial_overscan_above_trails = 4


class MockInstance:
    parallel_species = 1
    parallel_ccd = 2
    serial_species = 3
    serial_ccd = 4
    hyper_noise_scalars = [5]


@pytest.fixture(name="parallel_hyper_analysis")
def make_parallel_hyper_analysis():
    phase = ph.ParallelHyperPhase(phase_name='test_phase')
    return phase.make_analysis([], [], results=[MockResult])


@pytest.fixture(name="serial_hyper_analysis")
def make_serial_hyper_analysis():
    phase = ph.SerialHyperPhase(phase_name='test_phase')
    return phase.make_analysis([], [], results=[MockResult])


@pytest.fixture(name="parallel_serial_hyper_analysis")
def make_parallel_serial_hyper_analysis():
    phase = ph.ParallelSerialHyperPhase(phase_name='test_phase')
    return phase.make_analysis([], [], results=[MockResult])


class TestHyperPhase(object):
    def test_types(self):
        parallel_phase = ph.ParallelHyperPhase(phase_name='test_phase')
        serial_phase = ph.SerialHyperPhase(phase_name='test_phase')
        parallel_serial_phase = ph.ParallelSerialHyperPhase(phase_name='test_phase')

        assert isinstance(parallel_phase, ph.ParallelPhase)
        assert isinstance(parallel_phase, ph.HyperPhase)

        assert isinstance(serial_phase, ph.SerialPhase)
        assert isinstance(serial_phase, ph.HyperPhase)

        assert isinstance(parallel_serial_phase, ph.ParallelSerialPhase)
        assert isinstance(parallel_serial_phase, ph.HyperPhase)

    def test_hyper_phase_make_analysis(self, parallel_hyper_analysis,
                                       serial_hyper_analysis,
                                       parallel_serial_hyper_analysis):
        assert isinstance(parallel_hyper_analysis, ph.HyperPhase.Analysis)
        assert isinstance(serial_hyper_analysis, ph.HyperPhase.Analysis)
        assert isinstance(parallel_serial_hyper_analysis, ph.HyperPhase.Analysis)

    def test_describe(self, parallel_hyper_analysis,
                      serial_hyper_analysis,
                      parallel_serial_hyper_analysis):
        assert """
Running CTI analysis for... 

Parallel CTI: 
Parallel Species:
1

Parallel CCD
2

Hyper Parameters:
5

""" == parallel_hyper_analysis.describe(MockInstance)

        assert """
Running CTI analysis for... 

Serial CTI: 
Serial Species:
3

Serial CCD
4

Hyper Parameters:
5

""" == serial_hyper_analysis.describe(MockInstance)

        assert """
Running CTI analysis for... 

Parallel CTI: 
Parallel Species:
1

Parallel CCD
2

Serial CTI: 
Serial Species:
3

Serial CCD
4

Hyper Parameters:
5

""" == parallel_serial_hyper_analysis.describe(MockInstance)

    def test__make_analysis(self, phase, ci_data, cti_settings):
        analysis = phase.make_analysis(ci_datas=[ci_data], cti_settings=cti_settings)
        assert analysis.last_results is None
        assert (analysis.ci_datas_extracted[0].image == ci_data.image).all()
        assert (analysis.ci_datas_extracted[0].noise_map == ci_data.noise_map).all()
        assert (analysis.ci_datas_full[0].image == ci_data.image).all()
        assert (analysis.ci_datas_full[0].noise_map == ci_data.noise_map).all()
        assert analysis.cti_settings == cti_settings

    def test_noise_scaling_map_extraction(self):
        noise_scaling_maps = ph.ParallelHyperPhase(phase_name='test_phase').noise_scaling_maps_from_result(MockResult)
        assert noise_scaling_maps == [1, 2]

        noise_scaling_maps = ph.SerialHyperPhase(phase_name='test_phase').noise_scaling_maps_from_result(MockResult)
        assert noise_scaling_maps == [1, 3]

        noise_scaling_maps = ph.ParallelSerialHyperPhase(phase_name='test_phase').noise_scaling_maps_from_result(
            MockResult)
        assert noise_scaling_maps == [1, 2, 3, 4]

    def test_hyper_phase(self):
        phase = ph.ParallelHyperPhase(phase_name='test_phase')
        assert len(phase.hyper_noise_scalars) == 2
        assert len(phase.variable.priors) == 2

        instance = phase.variable.instance_from_unit_vector([0.5, 0.5])

        assert list(instance.hyper_noise_scalars) == [5.0, 5.0]


class TestResult(object):
    def test__fits_for_instance__uses_ci_data_fit(self, ci_data, cti_settings):
        parallel_species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
        parallel_ccd = arctic_params.CCD(well_notch_depth=0.1, well_fill_alpha=0.5, well_fill_beta=0.5,
                                         well_fill_gamma=0.5)

        phase = ph.ParallelPhase(parallel_species=[parallel_species], parallel_ccd=parallel_ccd, columns=1,
                                 phase_name='test_phase')

        analysis = phase.make_analysis(ci_datas=[ci_data], cti_settings=cti_settings)
        instance = phase.constant

        fits = analysis.fits_of_ci_data_extracted_for_instance(instance=instance)
        assert fits[0].ci_pre_cti.shape == (3, 1)

        full_fits = analysis.fits_of_ci_data_full_for_instance(instance=instance)
        assert full_fits[0].ci_pre_cti.shape == (3, 3)

    def test__fit_figure_of_merit__matches_correct_fit_given_galaxy_profiles(self, ci_data, cti_settings):
        parallel_species = arctic_params.Species(trap_density=0.1, trap_lifetime=1.0)
        parallel_ccd = arctic_params.CCD(well_notch_depth=0.1, well_fill_alpha=0.5, well_fill_beta=0.5,
                                         well_fill_gamma=0.5)

        phase = ph.ParallelPhase(parallel_species=[parallel_species], parallel_ccd=parallel_ccd,
                                 phase_name='test_phase')

        analysis = phase.make_analysis(ci_datas=[ci_data], cti_settings=cti_settings)
        instance = phase.constant
        cti_params = ph.cti_params_for_instance(instance=instance)
        fit_figure_of_merit = analysis.fit(instance=instance)

        mask = phase.mask_function(shape=ci_data.image.shape)
        ci_datas_fit = [phase.extract_ci_data(data=d, mask=mask) for d, mask in zip([ci_data], [mask])]
        fit = ci_fit.CIFit(masked_ci_data=ci_datas_fit[0], cti_params=cti_params, cti_settings=cti_settings)

        assert fit.likelihood == fit_figure_of_merit

    def test__results_of_phase_are_available_as_properties(self, ci_data, cti_settings):
        phase = ph.ParallelPhase(optimizer_class=NLO,
                                 parallel_species=[prior_model.PriorModel(arctic_params.Species)],
                                 parallel_ccd=arctic_params.CCD, phase_name='test_phase')

        result = phase.run(ci_datas=[ci_data], cti_settings=cti_settings)

        assert hasattr(result, 'most_likely_extracted_fits')
        assert hasattr(result, 'most_likely_full_fits')
        assert hasattr(result, 'noise_scaling_maps_of_ci_regions')
        assert hasattr(result, 'noise_scaling_maps_of_parallel_trails')
        assert hasattr(result, 'noise_scaling_maps_of_serial_trails')
        assert hasattr(result, 'noise_scaling_maps_of_serial_overscan_above_trails')

    def test__parallel_phase__noise_scaling_maps_of_images_are_correct(self, ci_data, cti_settings):
        phase = ph.ParallelPhase(optimizer_class=NLO,
                                 parallel_species=[prior_model.PriorModel(arctic_params.Species)],
                                 parallel_ccd=arctic_params.CCD, phase_name='test_phase')

        # The ci_region is [0, 1, 0, 1], therefore by changing the image at 0,0 to 2.0 there will be a residual of 1.0,
        # which for a noise_map entry of 2.0 gives a chi squared of 0.25..

        ci_data.image[0, 0] = 2.0
        ci_data.noise_map[0, 0] = 2.0

        result = phase.run(ci_datas=[ci_data], cti_settings=cti_settings)

        assert (result.noise_scaling_maps_of_ci_regions[0] == np.array([[0.25, 0.0, 0.0],
                                                                        [0.0, 0.0, 0.0],
                                                                        [0.0, 0.0, 0.0]])).all()

        assert (result.noise_scaling_maps_of_parallel_trails[0] == np.zeros((3, 3))).all()
        assert (result.noise_scaling_maps_of_serial_trails[0] == np.zeros((3, 3))).all()
        assert (result.noise_scaling_maps_of_serial_overscan_above_trails[0] == np.zeros((3, 3))).all()



from autofit.tools import path_util
from autocti.charge_injection import ci_frame
from autocti.charge_injection import ci_data
from autocti.charge_injection import ci_pattern

import os

test_path = '{}/../'.format(os.path.dirname(os.path.realpath(__file__)))

def shape_from_ci_data_resolution(ci_data_resolution):

    if ci_data_resolution == 'patch':
        return (36, 36)
    elif ci_data_resolution == 'lowres':
        return (100, 100)
    elif ci_data_resolution == 'midres':
        return (300, 300)
    elif ci_data_resolution == 'highres':
        return (600, 600)
    else:
        raise ValueError('An invalid data resolution was entered - ', ci_data_resolution)

def ci_regions_from_ci_data_resolution(ci_data_resolution):

    if ci_data_resolution == 'patch':
        return [(1, 7, 3, 30), (17, 23, 3, 30)]
    elif ci_data_resolution == 'lowres':
        return [(10, 30, 10, 80), (60, 80, 10, 80)]
    elif ci_data_resolution == 'midres':
        return [(10, 40, 10, 280), (110, 140, 10, 280), (210, 240, 10, 280)]
    elif ci_data_resolution == 'highres':
        return [(10, 40, 10, 580), (110, 140, 10, 580), (210, 240, 10, 580),
                (310, 340, 10, 580), (410, 440, 10, 580), (510, 540, 10, 580)]
    else:
        raise ValueError('An invalid data resolution was entered - ', ci_data_resolution)

def frame_geometry_from_ci_data_resolution(ci_data_resolution):

    if ci_data_resolution == 'patch':
        return CIFrameGeometryIntegration.patch()
    elif ci_data_resolution == 'lowres':
        return CIFrameGeometryIntegration.low_res()
    elif ci_data_resolution == 'midres':
        return CIFrameGeometryIntegration.mid_res()
    elif ci_data_resolution == 'highres':
        return CIFrameGeometryIntegration.high_res()
    else:
        raise ValueError('An invalid data-type was entered when generating the test-data suite - ', ci_data_resolution)

def ci_data_resolution_from_shape(shape):

    if shape == (36, 36):
        return 'patch'
    elif shape == (100, 100):
        return 'lowres'
    elif shape == (300, 300):
        return 'midres'
    elif shape == (600, 600):
        return 'highres'
    else:
        raise ValueError('An invalid shape was entered when generating the data-type - ', shape)

def load_test_ci_data(ci_data_type, ci_data_model, ci_data_resolution, normalization, load_cosmic_ray_image=False):

    frame_geometry = frame_geometry_from_ci_data_resolution(ci_data_resolution=ci_data_resolution)
    ci_regions = ci_regions_from_ci_data_resolution(ci_data_resolution=ci_data_resolution)

    pattern = ci_pattern.CIPatternUniform(normalization=normalization, regions=ci_regions)
    data_path = path_util.make_and_return_path_from_path_and_folder_names(
        path=test_path, folder_names=['data', ci_data_type, ci_data_model, ci_data_resolution])

    normalization = str(int(pattern.normalization))

    if not load_cosmic_ray_image:
        return ci_data.ci_data_from_fits(frame_geometry=frame_geometry, ci_pattern=pattern,
                                         image_path=data_path + 'image_' + normalization + '.fits',
                                         noise_map_path=data_path + 'noise_map_' + normalization + '.fits',
                                         ci_pre_cti_path=data_path + 'ci_pre_cti_' + normalization + '.fits')
    elif load_cosmic_ray_image:
        return ci_data.ci_data_from_fits(frame_geometry=frame_geometry, ci_pattern=pattern,
                                         image_path=data_path + 'image_' + normalization + '.fits',
                                         noise_map_path=data_path + 'noise_map_' + normalization + '.fits',
                                         ci_pre_cti_path=data_path + 'ci_pre_cti_' + normalization + '.fits',
                                         cosmic_ray_image_path=data_path + 'cosmic_ray_image_' + normalization + '.fits')

class CIFrameGeometryIntegration(ci_frame.FrameGeometry):

    def __init__(self, corner, parallel_overscan, serial_prescan, serial_overscan):
        """This class represents the quadrant geometry of an integration quadrant."""
        super(CIFrameGeometryIntegration, self).__init__(corner=corner, parallel_overscan=parallel_overscan,
                                                         serial_prescan=serial_prescan, serial_overscan=serial_overscan)

    @classmethod
    def patch(cls):
        """This class represents the quadrant geometry of an integration quadrant."""
        return CIFrameGeometryIntegration(corner=(0, 0), parallel_overscan=ci_frame.Region((33, 36, 1, 30)),
                                          serial_overscan=ci_frame.Region((0, 33, 31, 36)),
                                          serial_prescan=ci_frame.Region((0, 36, 0, 1)))

    @classmethod
    def low_res(cls):
        """This class represents the quadrant geometry of an integration quadrant."""
        return CIFrameGeometryIntegration(corner=(0, 0), parallel_overscan=ci_frame.Region((90, 100, 10, 80)),
                                          serial_overscan=ci_frame.Region((0, 90, 80, 100)),
                                          serial_prescan=ci_frame.Region((0, 100, 0, 10)))

    @classmethod
    def mid_res(cls):
        """This class represents the quadrant geometry of an integration quadrant."""
        return CIFrameGeometryIntegration(corner=(0, 0), parallel_overscan=ci_frame.Region((280, 300, 10, 280)),
                                          serial_overscan=ci_frame.Region((0, 280, 280, 300)),
                                          serial_prescan=ci_frame.Region((0, 300, 0, 10)))

    @classmethod
    def high_res(cls):
        """This class represents the quadrant geometry of an integration quadrant."""
        return CIFrameGeometryIntegration(corner=(0, 0), parallel_overscan=ci_frame.Region((580, 600, 20, 580)),
                                          serial_overscan=ci_frame.Region((0, 580, 580, 600)),
                                          serial_prescan=ci_frame.Region((0, 600, 20, 1)))